@@ -1,8 +1,12 @@
 (* The following functions are implemented in perfcount.c *) 
 
-(* Returns true is we have the performance counters *)
+(* Returns true if we have the performance counters *)
 external has_performance_counters: unit -> bool = "has_performance_counters"
 
+(* Initializes the CPU speed and returns true if we have
+   the performance counters *)
+external reset_performance_counters: unit -> bool = "reset_performance_counters"
+
 (* Returns number of seconds since the first read *)
 external read_pentium_perfcount : unit -> float = "read_pentium_perfcount"
 
@@ -13,10 +17,18 @@ external sample_pentium_perfcount_20 : unit -> int = "sample_pentium_perfcount_2
 external sample_pentium_perfcount_10 : unit -> int = "sample_pentium_perfcount_10"
 
 
-(* Whether to use the performance counters (on Pentium only) *)
+(** Whether to use the performance counters (on Pentium only) *)
+type timerModeEnum =
+  | Disabled      (** Do not collect timing information *)
+  | SoftwareTimer (** Use OCaml's [Unix.time] for timing information *)
+  | HardwareTimer (** Use the Pentium's cycle counter to time code *)
+  | HardwareIfAvail (** Use the hardware cycle counter if availible; 
+                        otherwise use SoftwareTimer *)
 
-(* The performance counters are disabled by default. *)
-let do_use_performance_counters = ref false
+(* The performance counters are disabled by default.
+   This will always be one of Disabled | SoftwareTimer | HardwareTimer.
+   HardwareIfAvail is handled in reset. *)
+let timerMode = ref Disabled
 
                                         (* A hierarchy of timings *)
 
@@ -35,14 +47,22 @@ let top = { name = "TOTAL";
 let current : t list ref = ref [top]
 
 exception NoPerfCount
-let reset (perfcount: bool) = 
+let reset (mode: timerModeEnum) : unit = 
   top.sub <- [];
-  if perfcount then begin
-    if not (has_performance_counters ()) then begin
-      raise NoPerfCount
-    end
-  end;
-  do_use_performance_counters := perfcount
+  match mode with
+    Disabled
+  | SoftwareTimer -> timerMode := mode
+  | HardwareTimer -> 
+      if not (reset_performance_counters ()) then begin
+        timerMode := SoftwareTimer;
+        raise NoPerfCount
+      end;
+      timerMode := mode
+  | HardwareIfAvail ->
+      if (reset_performance_counters ()) then
+        timerMode := HardwareTimer
+      else
+        timerMode := SoftwareTimer
 
 
 
@@ -50,19 +70,20 @@ let print chn msg =
   (* Total up *)
   top.time <- List.fold_left (fun sum f -> sum +. f.time) 0.0 top.sub;
   let rec prTree ind node = 
-    if !do_use_performance_counters then 
-      (Printf.fprintf chn "%s%-20s          %8.5f s\n" 
+    if !timerMode = HardwareTimer then 
+      (Printf.fprintf chn "%s%-25s      %8.5f s\n" 
          (String.make ind ' ') node.name node.time)
     else
-      (Printf.fprintf chn "%s%-20s          %6.3f s\n" 
+      (Printf.fprintf chn "%s%-25s      %6.3f s\n" 
          (String.make ind ' ') node.name node.time);
 
-   List.iter (prTree (ind + 2)) node.sub
+   List.iter (prTree (ind + 2)) (List.rev node.sub)
   in
   Printf.fprintf chn "%s" msg; 
   List.iter (prTree 0) [ top ];
   Printf.fprintf chn "Timing used %s\n"
-    (if !do_use_performance_counters then "performance counters" else "Unix.time");
+    (if !timerMode = HardwareTimer then "Pentium performance counters"
+     else "Unix.time");
   let gc = Gc.quick_stat () in 
   let printM (w: float) : string = 
     Printf.sprintf "%.2fMb" (w *. 4.0 /. 1000000.0)
@@ -85,7 +106,7 @@ let print chn msg =
 
 (* Get the current time, in seconds *)
 let get_current_time () : float = 
-  if !do_use_performance_counters then 
+  if !timerMode = HardwareTimer then
     read_pentium_perfcount ()
   else
     (Unix.times ()).Unix.tms_utime
@@ -93,7 +114,7 @@ let get_current_time () : float =
 let repeattime limit str f arg = 
                                         (* Find the right stat *)
   let stat : t = 
-    let curr = match !current with h :: _ -> h | _ -> assert false in
+    let curr = match !current with h :: _ -> h | [] -> assert false in
     let rec loop = function
         h :: _ when h.name = str -> h
       | _ :: rest -> loop rest
@@ -121,7 +142,11 @@ let repeattime limit str f arg =
   repeatf 1
 
 
-let time str f arg = repeattime 0.0 str f arg
+let time str f arg =
+  if !timerMode = Disabled then
+    f arg
+  else
+    repeattime 0.0 str f arg
     
 
 let lastTime = ref 0.0
@@ -131,7 +156,18 @@ let timethis (f: 'a -> 'b) (arg: 'a) : 'b =
   lastTime := get_current_time () -. start; 
   res
   
-
+(** Return the cumulative time of all calls to {!Stats.time} and
+  {!Stats.repeattime} with the given label. *)
+(* Usually there will be only one occurence in the tree, but summing them all
+   makes more sense than choosing one arbitrarily *)
+let lookupTime (label:string) : float =
+  let time : float ref = ref 0.0 in
+  let rec search (x:t) : unit = 
+    if x.name = label then time := !time +. x.time;
+    List.iter search x.sub
+  in
+  search top;
+  !time
 
 
 
