@@ -281,13 +281,18 @@ module ForwardsDataFlow =
                     (docList (fun s -> num s.sid)) 
                     (List.rev
                        (Queue.fold (fun acc s -> s :: acc) [] worklist)));
-        try 
-          let s = Queue.take worklist in 
-          processStmt s;
-          fixedpoint ();
-        with Queue.Empty -> 
-          if !T.debug then 
-            ignore (E.log "FF(%s): done\n\n" T.name)
+        let keepgoing = 
+          try 
+            let s = Queue.take worklist in 
+            processStmt s;
+            true
+          with Queue.Empty -> 
+            if !T.debug then 
+              ignore (E.log "FF(%s): done\n\n" T.name);
+            false
+        in
+        if keepgoing then
+          fixedpoint ()
       in
       fixedpoint ()
           
@@ -317,6 +322,11 @@ module type BackwardsTransfer = sig
   (** For each block id, the data at the start. This data structure must be 
    * initialized with the initial data for each block *)
 
+  val funcExitData: t
+  (** The data at function exit.  Used for statements with no successors.
+      This is usually bottom, since we'll also use doStmt on Return 
+      statements. *)
+
   val combineStmtStartData: Cil.stmt -> old:t -> t -> t option
   (** When the analysis reaches the start of a block, combine the old data 
    * with the one we have just computed. Return None if the combination is 
@@ -355,8 +365,8 @@ module BackwardsDataFlow =
     let getStmtStartData (s: stmt) : T.t = 
       try IH.find T.stmtStartData s.sid
       with Not_found -> 
-        E.s (E.bug "BF(%s): stmtStartData is not initialized for %d"
-               T.name s.sid)
+        E.s (E.bug "BF(%s): stmtStartData is not initialized for %d: %a"
+               T.name s.sid d_stmt s)
 
     (** Process a statement and return true if the set of live return 
      * addresses on its entry has changed. *)
@@ -374,7 +384,7 @@ module BackwardsDataFlow =
              (* Do the default one. Combine the successors *)
              let res = 
                match s.succs with 
-                 [] -> E.s (E.bug "You must doStmt for the statements with no successors")
+                 [] -> T.funcExitData
                | fst :: rest -> 
                    List.fold_left (fun acc succ -> 
                      T.combineSuccessors acc (getStmtStartData succ))
@@ -437,28 +447,60 @@ module BackwardsDataFlow =
                     (docList (fun s -> num s.sid)) 
                     (List.rev
                        (Queue.fold (fun acc s -> s :: acc) [] worklist)));
-        try 
-          let s = Queue.take worklist in 
-          let changes = processStmt s in 
-          if changes then begin
-            (* We must add all predecessors of block b, only if not already 
-             * in and if the filter accepts them. *)
-            List.iter 
-              (fun p ->
-                if not (Queue.fold (fun exists s' -> exists || p.sid = s'.sid) 
-                          false worklist) &&
-                  T.filterStmt p s then 
-                  Queue.add p worklist)
-              s.preds;
-          end;
-          fixedpoint ();
+        let keepgoing = 
+          try 
+            let s = Queue.take worklist in 
+            let changes = processStmt s in 
+            if changes then begin
+              (* We must add all predecessors of block b, only if not already 
+               * in and if the filter accepts them. *)
+              List.iter 
+                (fun p ->
+                   if not (Queue.fold (fun exists s' -> exists || p.sid = s'.sid) 
+                             false worklist) &&
+                     T.filterStmt p s then 
+                       Queue.add p worklist)
+                s.preds;
+            end;
+            true
 
-        with Queue.Empty -> 
-          if !T.debug then 
-            ignore (E.log "BF(%s): done\n\n" T.name)
+          with Queue.Empty -> 
+            if !T.debug then 
+              ignore (E.log "BF(%s): done\n\n" T.name);
+            false
+        in
+        if keepgoing then
+          fixedpoint ();
       in
       fixedpoint ();
           
   end
 
 
+(** Helper utility that finds all of the statements of a function. 
+  It also lists the return statments (including statements that
+  fall through the end of a void function).  Useful when you need an
+  initial set of statements for BackwardsDataFlow.compute. *)
+let sink_stmts = ref []
+let all_stmts = ref []
+let sinkFinder = object(self)
+  inherit nopCilVisitor
+
+  method vstmt s =
+    all_stmts := s ::(!all_stmts);
+    match s.succs with
+      [] -> (sink_stmts := s :: (!sink_stmts);
+	     DoChildren)
+    | _ -> DoChildren
+
+end
+
+(* returns (all_stmts, return_stmts).   *)
+let find_stmts (fdec:fundec) : (stmt list * stmt list) =
+  ignore(visitCilFunction (sinkFinder) fdec);
+  let all = !all_stmts in
+  let ret = !sink_stmts in
+  all_stmts := [];
+  sink_stmts := [];
+  all, ret
+
