@@ -39,6 +39,7 @@
 (* tree visitor and rewriter for cabs *)
 
 open Cabs
+open Cabshelper
 open Trace
 open Pretty
 module E = Errormsg
@@ -92,7 +93,10 @@ class type cabsVisitor = object
   method vExitScope: unit -> unit
 end
     
-let visitorLocation = ref { filename = ""; lineno = -1; byteno = -1; }
+let visitorLocation = ref { filename = ""; 
+			    lineno = -1; 
+			    byteno = -1;
+                            ident = 0}
     
         (* a default visitor which does nothing to the tree *)
 class nopCabsVisitor : cabsVisitor = object
@@ -188,7 +192,7 @@ and childrenTypeSpecifier vis ts =
       if n' != n || eo' != eo then (n', eo') else input
     in
     let nel' = mapNoCopy doOneField nel in
-    if s' != s || nel' != nel then (s', nel) else input
+    if s' != s || nel' != nel then (s', nel') else input
   in
   match ts with
     Tstruct (n, Some fg, extraAttrs) ->
@@ -433,15 +437,23 @@ and childrenStatement vis s =
           let dl' = List.map (fun d' -> DEFINITION d') dl in
           BLOCK ({blabels = []; battrs = []; bstmts = dl' }, l)
     end
-  | ASM (sl, b, inl, outl, clobs, l) -> 
-      let childrenStringExp ((s, e) as input) = 
+  | ASM (sl, b, details, l) -> 
+      let childrenIdentStringExp ((i,s, e) as input) = 
         let e' = ve e in
-        if e' != e then (s, e') else input
+        if e' != e then (i,s, e') else input
       in
-      let inl' = mapNoCopy childrenStringExp inl in
-      let outl' = mapNoCopy childrenStringExp outl in
-      if inl' != inl || outl' != outl then 
-        ASM (sl, b, inl', outl', clobs, l) else s
+      let details' = match details with
+      | None -> details
+      | Some { aoutputs = outl; ainputs = inl; aclobbers = clobs } ->
+	  let outl' = mapNoCopy childrenIdentStringExp outl in
+	  let inl' = mapNoCopy childrenIdentStringExp inl in
+	  if outl' == outl && inl' == inl then
+	    details
+	  else
+	    Some { aoutputs = outl'; ainputs = inl'; aclobbers = clobs }
+      in
+      if details' != details then 
+        ASM (sl, b, details', l) else s
   | TRY_FINALLY (b1, b2, l) -> 
       let b1' = visitCabsBlock vis b1 in
       let b2' = visitCabsBlock vis b2 in
@@ -485,6 +497,9 @@ and childrenExpression vis e =
       let el' = mapNoCopy ve el in
       if el' != el then COMMA (el') else e
   | CONSTANT _ -> e
+  | PAREN e1 -> 
+      let e1' = ve e1 in
+      if e1' != e1 then PAREN (e1') else e 
   | VARIABLE s -> 
       let s' = vis#vvar s in
       if s' != s then VARIABLE s' else e
@@ -533,7 +548,7 @@ and childrenInitExpression vis ie =
     | ATINDEXRANGE_INIT (e1, e2) -> 
         let e1' = visitCabsExpression vis e1 in
         let e2' = visitCabsExpression vis e2 in
-        if e1' != e1 || e2' != e2 then ATINDEXRANGE_INIT (e1, e2) else iw
+        if e1' != e1 || e2' != e2 then ATINDEXRANGE_INIT (e1', e2') else iw
   in
   match ie with 
     NO_INIT -> ie
