@@ -55,9 +55,11 @@ let close_output _ =
 
 let set_output filename =
   close_output ();
-  (try out := Some (open_out filename)
-  with (Sys_error msg) ->
-    output_string stderr ("Error while opening output: " ^ msg); exit 1);
+  let out_chan = try open_out filename 
+    with Sys_error msg -> 
+    (output_string stderr ("Error while opening output: " ^ msg); exit 1) in
+  out := Some out_chan;
+  Whitetrack.setOutput out_chan;
   close_me := true
 
    (* Signal that we are in MS VC mode *)
@@ -84,20 +86,24 @@ begin
   | None -> true
 end
 
+let printNotice = ref false
+
 (*
 ** Argument definition
 *)
 let args : (string * Arg.spec * string) list =
 [
-  "--cabsonly", Arg.String set_output, "<fname>: CABS output file name";
+  "--cabsonly", Arg.String set_output, "<fname> CABS output file name";
   "--printComments", Arg.Unit (fun _ -> Cprint.printComments := true),
-             ": print cabs tree structure in comments in cabs output";
+             " print cabs tree structure in comments in cabs output";
   "--patchFile", Arg.String (fun pf -> patchFileName := pf),
-             "<fname>: name the file containing patching transformations";
+             "<fname> name the file containing patching transformations";
   "--printPatched", Arg.Unit (fun _ -> printPatchedFiles := true),
-             ": print patched CABS files after patching, to *.patched";
+             " print patched CABS files after patching, to *.patched";
   "--printProtos", Arg.Unit (fun _ -> doPrintProtos := true),
-             ": print prototypes to safec.proto.h after parsing";
+             " print prototypes to safec.proto.h after parsing";
+  "--printNotice", Arg.Set printNotice,
+             " include a comment saying printed by FrontC";
 ]
 
 exception ParseError of string
@@ -154,7 +160,7 @@ begin
   (match !out with
     Some o -> begin
       (trace "sm" (dprintf "writing the cabs output\n"));
-      output_string o ("/* Generated by Frontc */\n");
+      if !printNotice then output_string o ("/* Generated by Frontc */\n");
       Stats.time "printCABS" (Cprint.printFile o) patched;
       close_output ();
       raise CabsOnly
@@ -167,6 +173,14 @@ begin
   patched
 end
 
+and clexer lexbuf =
+    Clexer.clear_white ();
+    Clexer.clear_lexeme ();
+    let token = Clexer.initial lexbuf in
+    let white = Clexer.get_white () in
+    let cabsloc = Clexer.currentLoc () in
+    let lexeme = Clexer.get_extra_lexeme () ^ Lexing.lexeme lexbuf in
+    white,lexeme,token,cabsloc
 
 (* just parse *)
 and parse_to_cabs_inner (fname : string) =
@@ -175,7 +189,8 @@ and parse_to_cabs_inner (fname : string) =
     flush !E.logChannel;
     E.hadErrors := false;
     let lexbuf = Clexer.init fname in
-    let cabs = Stats.time "parse" (Cparser.file Clexer.initial) lexbuf in
+    let cabs = Stats.time "parse" (Cparser.interpret (Whitetrack.wraplexer clexer)) lexbuf in
+    Whitetrack.setFinalWhite (Clexer.get_white ());
     Clexer.finish ();
     (fname, cabs)
   with (Sys_error msg) -> begin
@@ -237,20 +252,16 @@ end
 
 
 
-let parse fname =
+let parse_helper fname =
   (trace "sm" (dprintf "parsing %s to Cabs\n" fname));
   let cabs = parse_to_cabs fname in
   (* Now (return a function that will) convert to CIL *)
   fun _ ->
     (trace "sm" (dprintf "converting %s from Cabs to CIL\n" fname));
-    let cil = Stats.time "conv" Cabs2cil.convFile cabs in
+    let cil = Stats.time "convert to CIL" Cabs2cil.convFile cabs in
     if !doPrintProtos then (printPrototypes cabs);
-    cil
-
-
-
-
-
-
+    cabs, cil
 
+let parse fname = (fun () -> snd(parse_helper fname ()))
 
+let parse_with_cabs fname = (fun () -> parse_helper fname ())
