@@ -688,9 +688,7 @@ and matchEnumInfo (oldfidx: int) (oldei: enuminfo)
   else begin
     (* Replace with the representative data *)
     let oldei = oldeinode.ndata in
-    let oldfidx = oldeinode.nfidx in
     let ei = einode.ndata in
-    let fidx = einode.nfidx in
     (* Try to match them. But if you cannot just make them both integers *)
     try
       (* We do not have a mapping. They better be defined in the same way *)
@@ -1345,7 +1343,7 @@ let oneFilePass2 (f: file) =
             mergePushGlobals (visitCilGlobal renameVisitor g)
           end
 
-      | GVar (vi, init, l) as g ->
+      | GVar (vi, init, l) ->
           currentLoc := l;
           incr currentDeclIdx;
           let vi' = processVarinfo vi l in
@@ -1358,22 +1356,24 @@ let oneFilePass2 (f: file) =
               let prevVar, prevInitOpt, prevLoc =
                 (H.find emittedVarDefn vi'.vname) in
               (* previously defined; same initializer? *)
-              if (equalInitOpts prevInitOpt init.init) then (
+              if (equalInitOpts prevInitOpt init.init)
+                || (init.init = None) then (
                 (trace "mergeGlob"
                   (P.dprintf "dropping global var %s at %a in favor of the one at %a\n"
                              vi'.vname  d_loc l  d_loc prevLoc));
                 false  (* do not emit *)
               )
-              else (
-                (ignore (warn "global var %s at %a has different initializer than %a\n"
-                              vi'.vname  d_loc l  d_loc prevLoc));
-                (* emit it so we get a compiler error.. I think it would be
-                 * better to give an error message and *not* emit, since doing
-                 * this explicitly violates the CIL invariant of only one GVar
-                 * per name, but the rest of this file is very permissive so
-                 * I'll be similarly permissive.. *)
+              else if prevInitOpt = None then (
+                (* We have an initializer, but the previous one didn't.
+                   We should really convert the previous global from GVar
+                   to GVarDecl, but that's not convenient to do here. *)
                 true
               )
+              else ( 
+                (* Both GVars have initializers. *)
+                (E.s (error "global var %s at %a has different initializer than %a\n"
+                              vi'.vname  d_loc l  d_loc prevLoc));
+              )
             with Not_found -> (
               (* no previous definition *)
               (H.add emittedVarDefn vi'.vname (vi', init.init, l));
@@ -1432,7 +1432,6 @@ let oneFilePass2 (f: file) =
               (* If we must do alpha conversion then temporarily set the 
                * names of the local variables and formals in a standard way *)
               let nameId = ref 0 in 
-              let newName () = incr nameId;  in
               let oldNames : string list ref = ref [] in
               let renameOne (v: varinfo) = 
                 oldNames := v.vname :: !oldNames; 
@@ -1751,7 +1750,7 @@ let merge (files: file list) (newname: string) : file =
     { fileName = newname;
       globals  = revonto (revonto [] !theFile) !theFileTypes;
       globinit = None;
-      globinitcalled = false } in
+      globinitcalled = false;} in
   init (); (* Make the GC happy *)
   (* We have made many renaming changes and sometimes we have just guessed a 
    * name wrong. Make sure now that the local names are unique. *)
